/*
Copyright (C) 2006-2007 by Carlos H. P. da Silva <carloshp.silva@gmail.com>.

This file is part of LameScript 2.0.

    LameScript is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/
// ====================================================================
// lscript2.c
// ====================================================================
// LameScript path string syntax:
//
// $script:section/key
//
// Ex: suppose we have a script called "myscript", and
// this script has a section called "mysection", which
// in turn has a key called "mykey". The full path string
// would be:
//
// "$myscript:mysection/mykey"
//
// You can also use only partial paths, like:
// "$myscript" -> defines only the current script
// or
// "$myscript:mysection" -> defines both script and section
// or
// ":mysection" -> assumes it's a section into the
// current script
// or
// "/mykey" -> assumes it's a key into the current section
// and script
// or even
// "" or <null> -> uses the current path, as long it's a valid pair
//
// Note that when you informs a path (either full or relative),
// the currscr, currsec and/or currkey variables are affected.
// If you supply a malformed path string, you'll receive a
// LS_BADPATH error.
// If you supply an insufficient path string and
// try to access data, you'll receive a LS_BADPATH error.
// If you supply an inexistent path, you'll receive one of
// the following errors: LS_NOSCRIPT, LS_NOSECTION or
// LS_NOKEY error.
//

#include "lsdefs.h"

int			lserror;

lshash_t	*scripts = NULL;

// these are set by the LS_setPath() function
lshash_t	*currscr = NULL;
lshash_t	*currsec = NULL;
lshash_t	*currkey = NULL;
// ====================================================================

/**
 * Simple debug
 **/

// #define	LS_DEBUG	1
void	LS_WriteLog (char *fmt, ...)
{
#ifdef	LS_DEBUG
	va_list		argptr;
	char		logbuf[2048];
	FILE		*logfile;

	va_start	(argptr, fmt);
	vsprintf	(logbuf, fmt, argptr);
	va_end		(argptr);
	if ((logfile = fopen("ls.log", "a+")) != NULL)
	{
		fprintf (logfile, "%s\n", logbuf);
		fclose	(logfile);
	}
#endif
}

int		LS_getStatus (void)
{
	return (lserror);
}

/**
 * Parses the path string
 * and sets the pointers
 **/
int		LS_setPath (char *fmt, ...)
{
	int		i = 0, d = 0;
	char	buf[MAXPATHSIZE];
	char	path[MAXPATHSIZE];
	va_list	argptr;

	if (!scripts)
	{
		lserror = LS_NOINIT;
		return -1;
	}

	va_start (argptr, fmt);
	vsprintf (path, fmt, argptr);
	va_end (argptr);
	lserror = LSOK;
	memset (buf, '\0', MAXPATHSIZE);
	while ((i < MAXPATHSIZE) && (path[i] != '\0'))
	{
		switch (path[i])
		{
			case	'$':
				i++;
				d = 0;
				while ((path[i] != '\0') && (path[i] != ':') && (path[i] != '/') && (i < MAXPATHSIZE))
				{
					buf[d++] = path[i++];
				}

				buf[d] = '\0';
				currscr = (lshash_t *) _getHashTableItem (scripts, buf);
				currsec = NULL;
				currkey = NULL;
				if (!currscr)
				{
					lserror = LS_NOSCRIPT;
					return -1;
				}

				break;
			case	':':
				if (!currscr)
				{
					lserror = LS_BADPATH;
					return -1;
				}

				i++;
				d = 0;
				while ((path[i] != '\0') && (path[i] != ':') && (path[i] != '/') && (i < MAXPATHSIZE))
				{
					buf[d++] = path[i++];
				}

				buf[d] = '\0';
				currsec = (lshash_t *) _getHashTableItem (currscr, buf);
				currkey = NULL;
				if (!currsec)
				{
					lserror = LS_NOSECTION;
					return -1;
				}

				break;
			case	'/':
				if ((!currscr) || (!currsec))
				{
					lserror = LS_BADPATH;
					return -1;
				}

				i++;
				d = 0;
				while ((path[i] != '\0') && (path[i] != ':') && (path[i] != '/') && (i < MAXPATHSIZE))
				{
					buf[d++] = path[i++];
				}

				buf[d] = '\0';
				currkey = (lshash_t *) _getHashTableItem (currsec, buf);
				if (!currkey)
				{
					lserror = LS_NOKEY;
					return -1;
				}

				break;
			default:
				lserror = LS_BADPATH;
				return -1;
		}
	}

	return 0;
}

/**
 * Gets the current path.
 **/
char*	LS_getPath (void)
{
	char	*buf;
	// FIXME: possible buffer overflow here
	buf = malloc (1024);
	memset (buf, '\0', 1024);
	if (currscr)
	{
		sprintf (buf, "$%s", currscr->name);
	}

	if (currsec)
	{
		sprintf (buf, "%s:%s", buf, currsec->name);
	}

	if (currkey)
	{
		sprintf (buf, "%s/%s\0", buf, currkey->name);
	}

	return buf;
}

// ====================================================================
// script handling
int		LS_delScript (char *name)
{
    lshitem_t   *p;
    int         r;

    currscr = (lshash_t *) _getHashTableItem (scripts, name);
    if (!currscr)
    {
        lserror = LS_NOSCRIPT;
        return -1;
    }

    for (r = 0; r < HASHBUCKETS; r++)
    {
        p = (lshitem_t *) currscr->items[r];
        if ((p) && (p->name))
        {
            LS_delSection(name);
        }
    }

    _delHashTableItem (scripts, name);
	currscr = NULL;
	currsec = NULL;
	currkey = NULL;
    return 0;
}

/**
 * Creates a new script.
 * If already exists, removes
 * all its contents.
 * Also becomes the current
 * script.
 **/
int		LS_setScript (char *name)
{
	LS_delScript(name);
	currscr = _newHashTable (name);
	return (_setHashTableItem (scripts, name, currscr));
}

int		LS_getScriptCount (void)
{
	if (scripts)
	{
		return	scripts->count;
	}
	else
	{
		return -1;
	}
}
// ====================================================================
// section handling
int		LS_delSection (char *name)
{
    lshitem_t   *p;
    int         r;

    if (!currscr)
    {
        lserror = LS_NOSCRIPT;
        return -1;
    }

    currsec = (lshash_t *) _getHashTableItem (currscr, name);
    if (!currsec)
    {
        lserror = LS_NOSECTION;
        return -1;
    }

    for (r = 0; r < HASHBUCKETS; r++)
    {
        p = (lshitem_t *)currsec->items[r];
        if ((p) && (p->name))
        {
            LS_delKey(p->name);
        }
    }

	_delHashTableItem (currscr, name);
	currsec = NULL;
	currkey = NULL;
	return 0;
}

int		LS_setSection (char *name)
{
	LS_delSection (name);
	currsec = _newHashTable (name);
	return (_setHashTableItem (currscr, name, currsec));
}

int		LS_getSectionCount (void)
{
	if (currscr)
	{
		return currscr->count;
	}
	else
	{
		return -1;
	}
}
// ====================================================================
// keys handling
/**
 * Removes a key and all
 * associated items
 **/
int		LS_delKey (char *name)
{
    if (!currsec)
    {
        lserror = LS_NOSECTION;
        return -1;
    }

    currkey = (lshash_t *) _getHashTableItem (currsec, name);
    if (!currkey)
    {
        lserror = LS_NOKEY;
        return -1;
    }

    _clearHashTable (currkey);
    _delHashTableItem (currsec, name);
    currkey = NULL;
    return 0;
}

/**
 * Creates a new key.
 * Removes if already exists.
 **/
int		LS_setKey (char *name)
{
	LS_delKey(name);
	currkey = _newHashTable(name);
	return (_setHashTableItem (currsec, name, currkey));
}

/**
 * Gets the item count for the key.
 **/
int		LS_getKeyCount (void)
{
	if (currsec)
	{
		return (_getHashTableCount (currsec));
	}
	else
	{
		return -1;
	}
}
// ====================================================================
// values handling

/**
 * Adds a value to the current key.
 * List is limited to 65536 items.
 **/
char	_itnmbuf[16];
void	LS_addValue (char *value)
{
	if (!currkey)
	{
		lserror = LS_NOKEY;
		return;
	}

	if (currkey->count > MAXVALNUM)
	{
		lserror = LS_FULLKEY;
		return;
	}

	/*
		The list size limit is actually a hack,
		related to the current iterator API
		implementation. This may (or not) change
		in future versions.
	*/
	sprintf (_itnmbuf, "ITEM_%04X\0", currkey->count);
	if (_setHashTableItem (currkey, _itnmbuf, strdup(value)) != -1)
	{
		lserror = LSOK;
	}
	else
	{
		lserror = LS_BADHASHCODE;
	}
}

/**
 * Sets a specific position into
 * the current key value list.
 * Position must be a value between
 * 0 and 65535.
 **/
void	LS_setValue (char *value, int position)
{
	if (!currkey)
	{
		lserror = LS_NOKEY;
		return;
	}

	if ((position < 0) || (position > MAXVALNUM-1))
	{
		lserror = LS_BADKEYINDEX;
		return;
	}

	sprintf (_itnmbuf, "ITEM_%04X\0", position);
	_setHashTableItem (currkey, _itnmbuf, value);
}

/**
 * Returns the item from the
 * value list, or null in
 * case of any error
 **/
// Better performance and avoid
// memory leaks during intensive
// use (such inside loops)
char	_lsbuf[MAXVALSIZE+1];
char	*LS_getValue (int position)
{
	if (!currkey)
	{
		lserror = LS_NOKEY;
		return NULL;
	}

	if ((position < 0) || (position > 65535))
	{
		lserror = LS_BADKEYINDEX;
		return NULL;
	}

	sprintf (_itnmbuf, "ITEM_%04X\0", position);
	sprintf (_lsbuf, "%s\0", _getHashTableItem (currkey, _itnmbuf));
	return _lsbuf;
}

/**
 * Sets an iterator from
 * the desired position
 **/
int		LS_getIterator (int position)
{
	if (!currkey)
	{
		lserror = LS_NOKEY;
		return -1;
	}

	if ((position < 0) || (position > 65535))
	{
		lserror = LS_BADKEYINDEX;
		return -1;
	}

	sprintf (_itnmbuf, "ITEM_%04X\0", position);
	return (_setHashTableIterator (currkey, _itnmbuf));
}

/**
 * Checks if there are
 * more items in the
 * value list
 **/
int		LS_hasNext (void)
{
	return (_hasNextHashTableItem());
}

/**
 * Returns the next item
 * from the value list
 **/
char	*LS_nextValue (void)
{
	return strdup((char *)_nextHashTableItem ());
}

/**
 * Returns the total of
 * items in the value list
 **/
int		LS_getValueCount (void)
{
	return (_getHashTableCount (currkey));
}
// ====================================================================
// I/O functions
//
int		LS_saveScriptFile (FILE *fout)
{
	int			iSections, iKeys;
	int			cSections, cKeys;
	lshitem_t	*ptrKey, *ptrSec;

	if (fout)
	{
		if (currscr)
		{
			fprintf (fout, "# Generated by LameScript %d.%02d, modify carefully\n", LSVERSION, LSRELEASE);
			fprintf (fout, "# script: %s\n", currscr->name);
			cSections = 0;
			iSections = 0;

			while ((cSections < currscr->count) && (iSections < HASHBUCKETS))
			{
				ptrSec = currscr->items[iSections];
				while ((ptrSec) && (ptrSec->name))
				{
					currsec = (lshash_t *) ptrSec->item;
					if (currsec)
					{
						fprintf (fout, "\n%s {\n", currsec->name);
						cKeys = 0;
						iKeys = 0;
						while ((cKeys < currsec->count) && (iKeys < HASHBUCKETS))
						{
							ptrKey = currsec->items[iKeys];
							while ((ptrKey) && (ptrKey->name))
							{
								currkey = (lshash_t *) ptrKey->item;
								// DEBUG
								// fprintf (fout, "// %d value(s)\n", currkey->count);
								fprintf (fout, "\t%s=", currkey->name);
								LS_getIterator (0);
								while (LS_hasNext())
								{
									fprintf (fout, "\"%s\"", LS_nextValue());
									fprintf (fout, "%s", (LS_hasNext() ? "," : ";\n"));
								}

								cKeys++;
								ptrKey = (lshitem_t *)ptrKey->next;
							}

							iKeys++;
						}

						fprintf (fout, "}\n");
						cSections++;
						ptrSec = (lshitem_t *)ptrSec->next;
					}
				}

				iSections++;
			}
		}

		fclose (fout);
		return 0;
	}

	return -1;
}

int		LS_saveScript (char *filename)
{
	FILE	*fout;

	if ((fout = fopen(filename, "wt")) != NULL)
	{
		return LS_saveScriptFile (fout);
	}

	return -1;
}

/**
 * Parses and loads the script from file.
 * TODO: change error code to more specific errors.
 * TODO: return the bad script line ?
 **/
int		LS_loadScriptFile (FILE *fin, char *alias)
{
	char	buf[MAXVALSIZE];
	char	c = '\0', d;
	int		lastToken = LSTOK_BOF, count = 0, error = 0;

	LS_WriteLog("LS_loadScriptFile: start");
	if (fin)
	{
		fseek (fin, 0, SEEK_SET);
		LS_WriteLog("Calling LS_setScript(\"%s\")...", alias);
		LS_setScript (alias);
		LS_WriteLog("Entering main loop...");
		while ((!feof(fin)) && (!error))
		{
			fread (&c, sizeof(char), 1, fin);
			if (!c)
			{
				break;
			}

			switch (c)
			{
				case	'\t':
				case	' ':
				case	'\n':
				// just discard these
				break;

				// comments, can appear at any time
				case	'#':
					LS_WriteLog("found comment");
					while ((!feof(fin)) && (c != '\n'))
					{
						fread (&c, sizeof(char), 1, fin);
					}

					if (c != '\n')
					{
						// we reached the end of file without
						// reach the end of the script
						LS_WriteLog("1) lserror = LS_BADSCRIPT");
						lserror = LS_BADSCRIPT;
						error = 1;
						break;
					}

					LS_WriteLog("Comment discarded");
					lastToken = LSTOK_COMMENT;
				break;

				// value parsing
				case	'"':
					if ((lastToken != LSTOK_ENDKEY) && (lastToken != LSTOK_ENDVALUE))
					{
						LS_WriteLog("2) lserror = LS_BADSCRIPT");
						lserror = LS_BADSCRIPT;
						error = 1;
						break;
					}

					lastToken = LSTOK_STARTVALUE;
					memset (buf, '\0', MAXVALSIZE);
					count = 0;
					fread (&c, sizeof(char), 1, fin);
					while ((!feof(fin)) && (count < MAXVALSIZE - 1) && (c != '"'))
					{
                        // 2007-04-06
                        // added support for \n inside value strings
                        if (c == '\\')
                        {
                            fread (&d, sizeof(char), 1, fin);
                            if ((!feof(fin)) && (count < MAXVALSIZE - 1) && (d == 'n'))
                            {
                                buf[count++] = '\n';
                            }
                            else if ((!feof(fin)) && (count < MAXVALSIZE - 1) && (d != '"'))
                            {
                                buf[count++] = c;
                                buf[count++] = d;
                            }
                        }
                        else
                        {
						    buf[count++] = c;
                        }

  						fread (&c, sizeof(char), 1, fin);
					}

					if (c != '"')
					{
						// unexpected EOF or too big value string
						LS_WriteLog("3) lserror = LS_BADSCRIPT");
						lserror = LS_BADSCRIPT;
						error = 1;
						break;
					}

					LS_WriteLog("Found value \"%s\".", buf);
					lastToken = LSTOK_ENDVALUE;
					LS_addValue (buf);
					error = (lserror != LSOK);
				break;

				// value separators
				case	',':
				case	';':
					if (lastToken != LSTOK_ENDVALUE)
					{
						// bad placed comma
						LS_WriteLog("4) lserror = LS_BADSCRIPT");
						lserror = LS_BADSCRIPT;
						error = 1;
						break;
					}
				break;

				// section close
				case	'}':
					if (lastToken != LSTOK_ENDVALUE)
					{
						// malformed section or pair
						LS_WriteLog("5) lserror = LS_BADSCRIPT");
						lserror = LS_BADSCRIPT;
						error = 1;
						break;
					}
					LS_WriteLog("Found end of section");
					lastToken = LSTOK_ENDSECTION;
				break;
				default:
					memset (buf, '\0', MAXVALSIZE);
					count = 0;
					buf[count++] = c;
					if ((lastToken != LSTOK_BOF) &&
						(lastToken != LSTOK_COMMENT) &&
						(lastToken != LSTOK_ENDSECTION) &&
						(lastToken != LSTOK_STARTSECTION) &&
						(lastToken != LSTOK_ENDVALUE))
					{
						LS_WriteLog("6) lserror = LS_BADSCRIPT");
						lserror = LS_BADSCRIPT;
						error = 1;
						break;
					}

					fread (&c, sizeof(char), 1, fin);
					while ((!feof(fin)) && (count < MAXVALSIZE - 1))
					{
						if (c == '{')
						{
							if ((lastToken != LSTOK_BOF) &&
								(lastToken != LSTOK_COMMENT) &&
								(lastToken != LSTOK_ENDSECTION))
							{
								// invalid section start
								LS_WriteLog("7) lserror = LS_BADSCRIPT");
								lserror = LS_BADSCRIPT;
								error = 1;
								break;
							}

							lastToken = LSTOK_STARTSECTION;
							LS_WriteLog("Found start of section");
							break;
						}

						if (c == '=')
						{
							if ((lastToken != LSTOK_STARTSECTION) &&
								(lastToken != LSTOK_COMMENT) &&
								(lastToken != LSTOK_ENDVALUE))
							{
								// invalid key start
								LS_WriteLog("8) lserror = LS_BADSCRIPT");
								lserror = LS_BADSCRIPT;
								error = 1;
								break;
							}

							lastToken = LSTOK_STARTKEY;
							LS_WriteLog("Found key start");
							break;
						}

						if ((c != '\t') && (c != ' ') && (c != '\n'))
						{
							buf[count++] = c;
						}

						fread (&c, sizeof(char), 1, fin);
					}

					if (lastToken == LSTOK_STARTKEY)
					{
						LS_WriteLog("Found key \"%s\"", buf);
						LS_setKey (buf);
						lastToken = LSTOK_ENDKEY;
						break;
					}

					if (lastToken == LSTOK_STARTSECTION)
					{
						LS_WriteLog("Found section \"%s\"", buf);
						LS_setSection (buf);
						break;
					}

					// malformed script
					LS_WriteLog("9) lserror = LS_BADSCRIPT");
					lserror = LS_BADSCRIPT;
					error = 1;
				break;
			}
		}

		if ((error) ||
			(lserror != LSOK) ||
			((lastToken != LSTOK_COMMENT) && (lastToken != LSTOK_ENDSECTION)))
		{
			LS_WriteLog("ABNORMAL END");
			return -1;
		}

		return 0;
	}

	return -1;
}

int		LS_loadScript (char *filename, char *alias)
{
	FILE	*fin= fopen (filename, "rt");
	int		result;

	LS_WriteLog ("LS_LoadScript(\"%s\", \"%s\"): start", filename, alias);
	if (fin)
	{
		LS_WriteLog("fin != NULL");
		result = LS_loadScriptFile (fin, alias);
		fclose(fin);
		LS_WriteLog("closed fin, returning...");
		return result;
	}

	return -1;
}

// Initialization and path functions
void    LS_clearAllScripts(void)
{
    lshitem_t   *p;
    int         r;

    for (r = 0; r < HASHBUCKETS; r++)
    {
        p = scripts->items[r];
        if ((p) && (p->name))
        {
            LS_delScript (p->name);
        }
    }
}

int		LS_init(void)
{
	if (!scripts)
	{
		scripts = _newHashTable ("scripts");
	}
	else
	{
		LS_clearAllScripts();
	}

	lserror = LSOK;
	return (scripts != (lshash_t *)NULL);
}

// 2007-08-04
// Added LS_hasValue()
// returns the item position
// or -1 if the informed value
// is null or not found

int		LS_hasValue (char *value)
{
	int l = 0;
	int c = 0;

	if (value)
	{
		l = strlen (value);
		if (LS_getIterator(0) == LSOK)
		{
			while (LS_hasNext())
			{
				if (!strnicmp (value, LS_nextValue(), l))
				{
					return (c);
				}

				c++;
			}
		}
	}

	return -1;
}