/*
Copyright (C) 2006-2007 by Carlos H. P. da Silva <carloshp.silva@gmail.com>.

This file is part of LameScript 2.0.

    LameScript is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/
// ====================================================================
// lshash.c
// This is a hash table / linked list combo component, inspired in
// the Java collection API (specifically, this code mimics the
// ArrayList class behaviour). Thanks to the linked list, you can insert
// an undefined number of items into the lshash_t structure, and thanks
// to the hash table is quickier to retrieve these items than it
// would be using a single big linked list. Also there's support to a
// single iterator for fast loops into the linked lists. However, the
// current implementation allows only one iterator at time. Thus, the
// use of the iterator API specifically is discouraged since it may (and
// likely will) change in future versions.
// ====================================================================
#include "lshash.h"

// TODO: support multiple iterators ?
lshitem_t	*iterator = NULL;

/**
 * Calculates the hash code
 * This method gives a good spread
 * along all buckets available
 **/
int			_getHashCode (char *name)
{
	int		middle, result = 0;

	if ((!name) || (!strlen(name)))
	{
		// bad name, no hash code for you!
		return -1;
	}

	middle = strlen(name) / 2;
	result	|= ((name[0] & 7) << 3);
	result	|= ((name[middle] & 7));
	return result;
}

/**
 * Creates an empty hash table
 **/
lshash_t	*_newHashTable (char *name)
{
	lshash_t	*hash = (lshash_t *) NULL;
	int			r;

	// let's get a bit paranoic here
	if ((name) && (strlen(name)))
	{
		hash = malloc (sizeof(lshash_t));
		// let's keep the paranoic mode on...
		if (hash)
		{
			hash->count = 0;
			// hash->items = malloc (sizeof(lshitem_t) * HASHBUCKETS);
			for (r = 0; r < HASHBUCKETS; r++)
			{
				hash->items[r] = malloc (sizeof(lshitem_t));
				hash->items[r]->item = (void *)NULL;
				hash->items[r]->name = (char *)NULL;
				hash->items[r]->next = (void *)NULL;
			}

			hash->name = malloc (strlen(name) + 1);
			sprintf(hash->name, "%s\0", name);
		}
	}

	return hash;
}

/**
 * Adds an item to the hash table
 * Returns the item hash code, or
 * -1 if failed.
 * Replaces the item content if
 * already exists.
 **/
int			_setHashTableItem (lshash_t *hash, char *name, void *item)
{
	int		hashcode = _getHashCode(name), size;
	lshitem_t	*p;

	if ((hash) && (hashcode != -1))
	{
		size = strlen(name);
		p = hash->items[hashcode];
		// walk along the chain list
		while ((p) && (p->name))
		{
			// ops, looks like it already exists
			if (!strnicmp(name, p->name, size))
			{
				p->item = item;
				return hashcode;
			}

			p = (lshitem_t *)p->next;
		}

		// ok, found the end
		p->item = item;
        p->name = strdup (name);
		p->next = malloc(sizeof(lshitem_t));
		p = (lshitem_t *)p->next;
		p->item = (void *)NULL;
		p->name = (char *)NULL;
		p->next = (void *)NULL;
		hash->count++;
	}
	else
	{
		hashcode = -1;
	}

	return hashcode;
}

/**
 * Returns a pointer to the
 * given item name or NULL
 * if not found.
 **/
void		*_getHashTableItem (lshash_t *hash, char *name)
{
	int		hashcode = _getHashCode(name), size;
	lshitem_t	*p;

	if ((hash) && (hashcode != -1))
	{
		size = strlen(name);
		p = hash->items[hashcode];
		// walk along the chain list until find the end or the item
		while ((p) && (p->name))
		{
			// gotcha!
			if (!strnicmp(name, p->name, size))
			{
				return p->item;
			}

			p = (lshitem_t *)p->next;
		}
	}

	// hmmm, it's not here...
	return NULL;
}

/**
 * Removes the named item
 * Returns the hash code or
 * -1 if not found
 **/
int			_delHashTableItem (lshash_t *hash, char *name)
{
	int		hashcode = _getHashCode(name), size, gotcha = 0;
	lshitem_t	*p, *p2=(lshitem_t *) NULL;

	if ((hash) && (hashcode != -1))
	{
		size = strlen(name);
		p = hash->items[hashcode];
		while ((p) && (p->name))
		{
			if (!strnicmp(name, p->name, size))
			{
				if ((p->name != (char *)NULL) && (p->item != (void *)NULL))
				{
                    p->item = (void *)NULL;
                    p->name = (char *)NULL;
				}

				p2 = (lshitem_t *) p->next;
                if (p2)
                {
				    p->name = p2->name;
    				p->item = p2->item;
				    p->next = p2->next;
                }

				hash->count--;
				gotcha = 1;
				break;
			}

            p = (lshitem_t *) p->next;
		}
	}

	if (gotcha)
	{
		return hashcode;
	}
	else
	{
		return -1;
	}
}

/**
 * Gets the total number of
 * items into the hash table
 **/
int			_getHashTableCount (lshash_t *hash)
{
	return hash->count;
}

/**
 * Removes all items from
 * the hash table
 **/
void		_clearHashTable (lshash_t *hash)
{
    lshitem_t   *p;
    int         r;

    while (hash->count)
    {
        for (r = 0; r < HASHBUCKETS; r++)
        {
            p = hash->items[r];
            if ((p) && (p->name))
            {
                _delHashTableItem (hash, p->name);
            }
        }
    }
}

// =============================================================
// Iterator functions, useful for quick loops through lists
// =============================================================
/**
 * Sets the iterator pointer
 * Returns the hash code or
 * -1 if failed
 **/
int			_setHashTableIterator (lshash_t *hash, char *name)
{
	int		hashcode = _getHashCode(name), gotcha = 0, size;

	if ((hash) && (hashcode != -1))
	{
		size = strlen (name);
		iterator = (lshitem_t *)hash->items[hashcode];
		while ((iterator) && (iterator->name) && (!gotcha))
		{
			if (!strnicmp (iterator->name, name, size))
			{
				gotcha = 1;
				break;
			}

			iterator = (lshitem_t *) iterator->next;
		}
	}

	if (!gotcha)
	{
		hashcode = -1;
	}

	return hashcode;
}

/**
 * Returns the next item
 * from the iterator list
 * Assumes _setHashTableIterator()
 * was previously executed
 **/
void		*_nextHashTableItem (void)
{
	void	*ptr = NULL;
	if (iterator != (lshitem_t *)NULL)
	{
		ptr = iterator->item;
		iterator = (lshitem_t *)iterator->next;
	}

	return ptr;
}

/**
 * Checks if there are items left to iterate
 **/
int			_hasNextHashTableItem (void)
{
	return ((iterator != NULL) && (iterator->name != NULL));
}